from .MNIST_base import MNIST_base
from .Omniglot_base import Omniglot_base
from .CIFAR10_base import CIFAR10_base
from .CIFAR100_base import CIFAR100_base
from .SVHN_base import SVHN_base
from .Madelon_base import Madelon_base
from .Moonway_base import Moonway_base

__all__ = [
    "MNIST_base",
    "Omniglot_base",
    "CIFAR10_base",
    "CIFAR100_base",
    "SVHN_base",
    "Madelon_base",
    "Moonway_base"
]
